#!/usr/bin/python
# -*- coding: utf-8 -*-
# Source Creator - Qt GUI for creating source files from a template
# Copyright (C) 2009 A. S. Budden
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

# Core modules:
import os
import sys
import imp
import re
import optparse
import logging

from PyQt4 import QtGui, QtCore

from SourceCreatorWindow import Ui_dlgSourceCreator
from Configuration import Configuration
from Settings import *

dlgConfiguration = None

TagMatch = '<\+(?P<keyword>.{1,}?)(?P<replace>;R)?\++>'
VarTagMatch = '<\+(?P<keyword>\$[A-Z]+\$)(?P<replace>;R)?\++>'
AskTagMatch = '#(?P<askstring>[A-Z_]{1,}?)#'
searchSpecial = '$^*[]\:'

TagRE = re.compile(TagMatch)
VarTagRE = re.compile(VarTagMatch)
AskTagRE = re.compile(AskTagMatch)

def GetStandardVariableDefinition(filename, variable):
	filename = os.path.basename(filename)
	basename = os.path.splitext(filename)[0]

	if variable == "$FILENAME$":
		return filename
	elif variable in ["$BASENAME$", "$MCPREFIX$"]:
		return basename
	elif variable in ["$UBASENAME$", "$UCPREFIX$"]:
		return basename.upper()
	elif variable == "$LBASENAME$":
		return basename.lower()
	elif variable == "$YEAR$":
		import time
		return time.strftime("%Y")
	elif variable == "$DATE$":
		import time
		return time.strftime("%d/%m/%Y")
	elif variable == "$DATETIME$":
		import time
		return time.strftime("%d/%m/%Y %H.%M.%S")
	else:
		sys.stderr.write("No match: %s" % variable)
		return ""



# Example header format (more than three # symbols is comment)
### TEMPLATE: Template Name
### CONFIG: PROJECT_TITLE;Project Title;;REMEMBER
### CONFIG: MODULE_TITLE;Module Title;;
### CONFIG: COMPILER_NAME;Compiler Name;Keil uVision;REMEMBER
### CONFIG: TARGET_TYPE;Target Description;STM32F10x;REMEMBER
### OPTION: DOXYSTRIP;REMEMBER
### BEGIN:FILE:c


def ReadTemplate(filename):
	fh = open(filename, 'r')

	FileDetails = \
			{
				'Name': '',
				'Contents': {},
				'Config': [],
				'Options': [],
			}

	InFile = False
	EndLine = '### END:FILE:'
	Extension = ''

	for line in fh:
		if InFile:
			if line.strip() == (EndLine + Extension):
				InFile = False
			else:
				FileDetails['Contents'][Extension] += line

		elif line.startswith('####'):
			pass

		elif line.startswith('### CONFIG: '):
			Components = line[len('### CONFIG:'):].strip().split(';')
			while len(Components) < 4:
				Components.append('')
			config = \
					{
						'Name':    Components[1],
						'Pattern': Components[0],
						'Default': Components[2],
						'Remember': False
					}

			if Components[3] == 'REMEMBER':
				config['Remember'] = True

			FileDetails['Config'].append(config)

		elif line.startswith('### OPTION: '):
			Components = line[len('### CONFIG:'):].strip().split(';')
			while len(Components) < 3:
				Components.append('')
			option = \
					{
						'Name': Components[0],
						'Default': False,
						'Remember': False
					}
			
			if Components[1].lower() == 'true':
				option['Default'] = True
			
			if Components[2] == 'REMEMBER':
				option['Remember'] = True

			FileDetails['Options'].append(option)

		elif line.startswith('### TEMPLATE: '):
			FileDetails['Name'] = line[len('### TEMPLATE:'):].strip()

		elif line.startswith('### BEGIN:FILE:'):
			Extension = line[len('### BEGIN:FILE:'):].strip()
			FileDetails['Contents'][Extension] = ''
			InFile = True

			
	fh.close()

	return FileDetails

def GetTemplates(template_dir):
	import glob
	template_list = glob.glob(os.path.join(template_dir, '*.tmpl'))
	return [ReadTemplate(i) for i in template_list]


class SourceCreator(QtGui.QDialog):
	def __init__(self, title=None,
			dir=None,
			default_size=None,
			dialog=True,
			maindir=None,
			parent=None):

		QtGui.QDialog.__init__(self)

		self.settings = QtCore.QSettings("gataki.co.uk", "Source Creator")

		self.ui = Ui_dlgSourceCreator()
		self.ui.setupUi(self)

		try:
			self.ui.listTemplates.currentIndexChanged[int].connect(self.ShowConfiguration)
			self.ui.buttonBox.accepted.connect(self.OnAccept)
			self.ui.txtBaseName.textChanged.connect(self.Validate)
			self.ui.tblConfig.itemChanged.connect(self.Validate)
			self.ui.btnBrowse.clicked.connect(self.OnBrowse)
			self.ui.btnConfigure.clicked.connect(self.ConfigureApplication)
		except AttributeError:
			QtCore.QObject.connect(self.ui.listTemplates, QtCore.SIGNAL("currentIndexChanged(int)"),
					self.ShowConfiguration)
			QtCore.QObject.connect(self.ui.buttonBox, QtCore.SIGNAL("accepted()"),
					self.OnAccept)
			QtCore.QObject.connect(self.ui.txtBaseName, QtCore.SIGNAL("textChanged(QString)"),
					self.Validate)
			QtCore.QObject.connect(self.ui.tblConfig, QtCore.SIGNAL("itemChanged(QTableWidgetItem*)"),
					self.Validate)
			QtCore.QObject.connect(self.ui.btnBrowse, QtCore.SIGNAL('clicked()'),
					self.OnBrowse)
			QtCore.QObject.connect(self.ui.btnConfigure, QtCore.SIGNAL('clicked()'),
					self.ConfigureApplication)

		if maindir is None:
			maindir = GetMainDir()

		self.maindir = maindir

		self.ui.txtWorkingDir.setReadOnly(True)
		self.ui.tblConfig.horizontalHeader().setStretchLastSection(True)

		if title:
			self.setWindowTitle(title)

		self.PopulateFromConfiguration()

	def PopulateFromConfiguration(self):
		templates_directory = ReadGlobalConfig('TemplatesDirectory',
				os.path.join(self.maindir, 'templates'),
				self.settings)

		self.source_dir = ReadGlobalConfig('SourceDirectory',
				'Source',
				self.settings)
		self.headers_dir = ReadGlobalConfig('HeadersDirectory',
				'Headers',
				self.settings)

		self.templates = GetTemplates(templates_directory)

		for template in sorted(self.templates):
			for ConfigItem in template['Config']:
				ConfigItem['Default'] = ReadConfig(ConfigItem, self.settings)
			for OptionItem in template['Options']:
				OptionItem['Default'] = ReadOption(OptionItem, self.settings)
		
		self.template_indices = {}
		for index in range(len(self.templates)):
			self.template_indices[self.templates[index]['Name']] = index

		self.ui.listTemplates.clear()
		for name in sorted(self.template_indices.keys()):
			self.ui.listTemplates.addItem(name)

		self.dir = os.getcwd()

		self.ui.txtWorkingDir.setText(self.dir)
		self.ui.chkUseSubDirs.setChecked(ReadGlobalOption('UseSubDirs',True,self.settings))

		defaultTemplate = ReadGlobalConfig('DefaultTemplate',None,self.settings)
		if defaultTemplate:
			index = self.ui.listTemplates.findText(defaultTemplate)
			if index != -1:
				self.ui.listTemplates.setCurrentIndex(index)

		self.Validate()

	def Validate(self):
		basename = str(self.ui.txtBaseName.text())
		ValidName = False

		if len(basename) == 0:
			warning = "Please enter file base name"
		elif not basename[0].isalpha() or not basename[0].isupper():
			warning = "Base name must start with an upper case letter"
		elif not basename.isalnum():
			warning = "Base name should only contain alphanumeric characters"
		elif basename.isupper():
			warning = "Upper case base name: macros and functions will appear identical"
		elif basename.islower():
			warning = "Lower case base name: naming convention suggests mixed case"
		else:
			warning = ""
			ValidName = True

		ValidEntries = True
		FirstMissingEntry = None
		for row in range(self.ui.tblConfig.rowCount()):
			item = self.ui.tblConfig.item(row, 1)
			if item:
				value = str(item.text())
				if len(value) == 0:
					ValidEntries = False
					FirstMissingEntry = str(self.ui.tblConfig.item(row, 0).text())

		if ValidName and FirstMissingEntry is not None:
			warning = "Missing configuration for '%s'" % FirstMissingEntry

		self.ui.lblWarning.setText(warning)

		self.ui.buttonBox.button(QtGui.QDialogButtonBox.Ok).setEnabled(ValidName and ValidEntries)

	def ConfigureApplication(self):
		self.dlgConfiguration = Configuration(self.settings)
		try:
			self.dlgConfiguration.accepted.connect(self.PopulateFromConfiguration)
		except AttributeError:
			QtCore.QObject.connect(self.dlgConfiguration, QtCore.SIGNAL('accepted()'),
					self.PopulateFromConfiguration)
		self.dlgConfiguration.show()
	
	def ShowConfiguration(self, index):

		template_name = str(self.ui.listTemplates.itemText(index))

		self.ui.tblConfig.clear()
		self.ui.tblConfig.setColumnCount(2)

		try:
			template = self.templates[self.template_indices[template_name]]
			self.ui.tblConfig.setRowCount(len(template['Config']))

			index = 0
			for index in range(len(template['Config'])):
				item = template['Config'][index]
				name = QtGui.QTableWidgetItem(item['Name'])
				default = QtGui.QTableWidgetItem(item['Default'])
				# Look at the 'remember' field here?
				# or maybe when loading the template
				name.setFlags(QtCore.Qt.ItemIsEnabled)
				default.setFlags(QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsSelectable)
				self.ui.tblConfig.setItem(index, 0, name)
				self.ui.tblConfig.setItem(index, 1, default)

			self.ui.chkStripDoxygenComments.setChecked(False)
			self.ui.chkStripDoxygenComments.setVisible(False)
			for item in template['Options']:
				if item['Name'] == 'DOXYSTRIP':
					self.ui.chkStripDoxygenComments.setVisible(True)
					if item['Default']:
						self.ui.chkStripDoxygenComments.setChecked(True)
		except KeyError:
			pass

	def OnBrowse(self):
		Directory = QtGui.QFileDialog.getExistingDirectory(
				self, "Select destination directory",
				self.ui.txtWorkingDir.text(),
				QtGui.QFileDialog.ShowDirsOnly | QtGui.QFileDialog.DontResolveSymlinks)
		if Directory:
			self.ui.txtWorkingDir.setText(Directory)

	def ParseContents(self, filename, original):
		template = self.templates[self.template_indices[str(self.ui.listTemplates.currentText())]]

		startposition = 0
		m = 1
		while m is not None:
			m = VarTagRE.search(original, startposition)
			if m is not None:
				keyword = m.group('keyword')
				replace = m.group('replace')
				variable = GetStandardVariableDefinition(filename, keyword)
				if variable != "":
					if replace is not None:
						width = len(m.group(0))
						variable = variable + " "*(width-len(variable))
				original = original.replace(m.group(0), variable, 1)

				startposition = m.start() + len(variable) + 1
		
		startposition = 0
		m = 1
		while m is not None:
			m = TagRE.search(original, startposition)
			if m is not None:
				keyword = m.group('keyword')
				replace = m.group('replace')

				am = AskTagRE.match(keyword)
				if am is not None:
					keyword = am.group('askstring')
					variable = ""
					for item in template['Config']:
						if keyword == item['Pattern']:
							variable = item['Value']
							break

					if variable != "":
						if replace is not None:
							width = len(m.group(0))
							variable = variable + " "*(width-len(variable))
					original = original.replace(m.group(0), variable, 1)

				startposition = m.start() + len(variable) + 1

		for index in range(len(template['Options'])):
			item = template['Options'][index]
			if item['Name'] == 'DOXYSTRIP':
				if item['Value'] is True:
					DoxyRE = re.compile(r'(/\*[*!].*?\*/\s*\n?|//[/!][^\n]*)', re.DOTALL)
					original = DoxyRE.sub('', original)

		return original


	def OnAccept(self):
		# Global options
		if self.ui.chkUseSubDirs.isChecked():
			WriteGlobalOption('UseSubDirs', True, self.settings)
		else:
			WriteGlobalOption('UseSubDirs', False, self.settings)

		template = self.templates[self.template_indices[str(self.ui.listTemplates.currentText())]]
		WriteGlobalConfig('DefaultTemplate', template['Name'], self.settings)

		# Store defaults
		for row in range(self.ui.tblConfig.rowCount()):
			name = str(self.ui.tblConfig.item(row, 0).text())
			value = str(self.ui.tblConfig.item(row, 1).text())

			for index in range(len(template['Config'])):
				item = template['Config'][index]
				if name == item['Name']:
					if value != item['Default'] and item['Remember']:
						WriteConfig(item, value, self.settings)
					template['Config'][index]['Value'] = value
					break


		for index in range(len(template['Options'])):
			item = template['Options'][index]
			if item['Name'] == 'DOXYSTRIP':
				item['Value'] = item['Default']
				if self.ui.chkStripDoxygenComments.isChecked() != item['Default']:
					item['Value'] = self.ui.chkStripDoxygenComments.isChecked()
					WriteOption(item, item['Value'], self.settings)

		# Now do the work - use the 'Value' items
		WorkingDir = str(self.ui.txtWorkingDir.text())
		BaseName = str(self.ui.txtBaseName.text())
		for extension in template['Contents'].keys():
			if self.ui.chkUseSubDirs.isChecked():
				if extension.startswith('c'):
					filename = os.path.join(WorkingDir,
							self.source_dir,
							BaseName + '.' + extension)
				elif extension.startswith('h'):
					filename = os.path.join(WorkingDir,
							self.headers_dir,
							BaseName + '.' + extension)
				else:
					filename = os.path.join(WorkingDir,
							BaseName + '.' + extension)
			else:
				filename = os.path.join(WorkingDir,
						BaseName + '.' + extension)

			if os.path.isdir(os.path.dirname(filename)):
				if os.path.exists(filename):
					QtGui.QMessageBox.critical(self, "Source Creator",
							"ERROR: File %s already exists" % filename)
					return
			else:
				os.makedirs(os.path.dirname(filename))

			fh = open(filename, 'w')

			contents = self.ParseContents(filename, template['Contents'][extension])
			if not contents.endswith('\n'):
				contents += '\n'

			fh.write(contents)

			fh.close()
		
if __name__ == "__main__":
	application = QtGui.QApplication(sys.argv)
	window = SourceCreator()
	window.show()
	application.exec_()

