#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Source Creator - Qt GUI for creating source files from a template
# Copyright (C) 2009 A. S. Budden
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

"""SourceCreator - Qt-based GUI for creating source files from a template

"""

import os
import sys
from bzrlib import errors, ui, registry
from bzrlib.option import Option
from bzrlib.commands import Command
import bzrlib.builtins

from bzrlib.lazy_import import lazy_import
from bzrlib.commands import register_command
lazy_import(globals(), '''
from PyQt4 import QtGui, QtCore

from bzrlib.plugins.qbzr.lib import i18n
from bzrlib.plugins.qbzr.lib.uifactory import QUIFactory
''')

CUR_DIR=u'.'

class PyQt4NotInstalled(errors.BzrError):

	_fmt = ('QBzr require at least PyQt 4.4 and '
			'Qt 4.4 to run. Please check your install')


def report_missing_pyqt(unbound):
	"""Decorator for q-commands run method to catch ImportError PyQt4
	and show explanation to user instead of scary traceback.
	See bugs: #240123, #163728
	"""
	def run(self, *args, **kwargs):
		try:
			return unbound(self, *args, **kwargs)
		except ImportError, e:
			if str(e).endswith('PyQt4'):
				raise PyQt4NotInstalled
			raise
	return run


def install_gettext(unbound):
	"""Decorator for q-commands run method to enable gettext translations."""
	def run(self, *args, **kwargs):
		i18n.install()
		try:
			return unbound(self, *args, **kwargs)
		finally:
			i18n.uninstall()
	return run


class QBzrCommand(Command):
	"""Base class for all q-commands.

	NOTES:
	1) q-command should define method '_qbzr_run' instead of 'run' (as in
	   bzrlib).
	2) The _qbzr_run method should return 0 for successfull exit
	   and 1 if operation was cancelled by user.
	3) The _qbzr_run method can return None, in this case return code will be
	   asked from self.main_window instance (if there is one).
	   self.main_window should be instance of QBzrWindow or QBzrDialog
	   with attribute "return_code" set to 0 or 1.
	"""
	
	@install_gettext
	@report_missing_pyqt
	def run(self, *args, **kwargs):
		# We must create the QApplication before we install the QUIFactory.
		# This is because once the factory is installed, it is possible that
		# opening a branch, etc will go to the ui_factory to request a
		# password, and the app must exist before the dialog can be created.
		self._application = QtGui.QApplication(sys.argv)
		ui.ui_factory = QUIFactory()
		
		# Set up global exception handling.
		from bzrlib.plugins.qbzr.lib.trace import excepthook
		sys.excepthook = excepthook
		
		try:
			ret_code = self._qbzr_run(*args, **kwargs)
			if ret_code is None:
				main_window = getattr(self, "main_window", None)
				if main_window is not None:
					ret_code = getattr(main_window, "return_code", None)
			return ret_code
		except Exception:
			ui_mode = ("ui_mode" in kwargs and kwargs["ui_mode"])
			from bzrlib.plugins.qbzr.lib.trace import report_exception
			return report_exception(ui_mode=ui_mode)



class cmd_qsourcecreator(QBzrCommand):
	"""Create source files from a template."""
	takes_args = []
	takes_options = []

	def _qbzr_run(self, ui_mode=False, directory=None):
		from SourceCreator import SourceCreator
		window = SourceCreator(maindir=os.path.dirname(__file__))
		window.show()
		self._application.exec_()


class LazyCommandProxy(registry._LazyObjectGetter):

	def __init__(self, module, name, aliases):
		super(LazyCommandProxy, self).__init__(module, name)
		self.aliases = aliases
		self.__name__ = name

	def __call__(self, *args, **kwargs):
		return self.get_obj()(*args, **kwargs)

	def __getattr__(self, name):
		return getattr(self.get_obj(), name)


def register_lazy_command(module, name, aliases, decorate=False):
	"""Lazily register a command.

	:param module: Name of the module where is the command defined
	:param name: Name of the command class; this Command subclass must
		exist in `module`
	:param aliases: List of command aliases
	:param decorate: If true, allow overriding an existing command
		of the same name; the old command is returned by this function.
		Otherwise it is an error to try to override an existing command.
	"""
	#try:
		## FIXME can't overwrite existing command
		#plugin_cmds.register_lazy(name, aliases, module)
	#except AttributeError:
	register_command(LazyCommandProxy(module, name, aliases), decorate)

register_lazy_command('bzrlib.plugins.source_creator', 'cmd_qsourcecreator', [])
